;; $Id: dispatch.scm,v 3.13 1994/10/05 14:05:32 queinnec Exp $ 
;;; Copyright (c) 1990-93 by Christian Queinnec. All rights reserved.

;;;                ***********************************************
;;;                    Small, Efficient and Innovative Class System
;;;                                    Meroon 
;;;                              Christian Queinnec  
;;;                   \'Ecole Polytechnique & INRIA--Rocquencourt
;;;                ************************************************

;;; This file contains the run-time functions that allows generic
;;; dispatching. Many functions are generic so one can add new kind of
;;; dispatchers. In that case, one has to extend the find-method
;;; function which is not generic since it must runs fast. 

;;; The predefined dispatchers classes appear in genesis.scm. Some
;;; types exist:
;;;   Immediate-Dispatcher does not check the class of the
;;;     discriminating variable and imposes the method to be found.
;;;   Subclass-Dispatcher tests the class of the discriminating variable 
;;;     and proposes two dispatchers whether the discriminating variable is
;;;     an instance or not of this class.
;;;   Indexed-Dispatcher finds a method in a vector indexed by the class-number
;;;     of the discriminating instance.
;;;   Global-Dispatcher uses a classes-wide Indexed-Dispatcher.
;;;   Linear-Dispatcher uses an Alist-like structure for multi-methods.

;;; Dispatchers in generic functions look like a decision tree that cannot
;;; be deeper than the following depth. When a method is added, the
;;; whole dispatcher is compressed to respect this depth. A zero depth
;;; means that all generic functions dispatch automatically with an
;;; Indexed-Dispatcher, a high number favours the elaboration of a
;;; decision tree which a high number of run-time tests. 
;;; A negative value is a special case meaning "dont try to compress
;;; dispatchers". It is used for debug purposes.

(define *dispatcher-max-depth* 
  (if-meroon-feature (not renumber-classes)
      -1    ; Use Global-Dispatchers
      3     ; compress dispatchers into less than 3 levels
      ) )

;;; extend the vector of classes as well as the dispatch tables of all
;;; generic functions (whether traced or not) that need it. This is
;;; necessary for indexed dispatchers that might need to be extended
;;; to take the new classes into account. Global-Dispatchers always
;;; have the size of vector *classes*. Whenever this latter grows, the
;;; formers grow similarly.

(define (extend-classes-number!)
  (set! *classes* (vector-extend *classes*))
  (let ((size (vector-length *classes*)))
    (sequence-map 
     (lambda (g)
       (when g 
         (let ((d (careless-Generic-dispatcher g)))
           (when (Global-Dispatcher? d)
             (let ((oldn (careless-Global-Dispatcher-method-length d))
                   (newd ((Class-allocator Global-Dispatcher-class) size)) )
               (do ((i 0 (+ i 1)))
                   ((>= i oldn))
                 (careless-set-Global-Dispatcher-method!
                  newd i (careless-Global-Dispatcher-method d i) ) )
               (careless-set-Generic-dispatcher! g newd) ) ) ) ) )
     *generics* )
    #f ) )

;; A method to fill holes in Indexed-Dispatcher. It is a sure sign of
;;; some internal error if it is ever invoked.

(define (barfing-method . arguments)
  (oo-apply report-meroon-error 'Domain 'barfing-method 
            "Absent method" arguments ) )

;;; Report that no method has been found for this GENERIC function and
;;; this set of ARGUMENTS. To increase code sharing, this is default
;;; method generator: see the expansion of define-generic.

(define (create-default-generic-method generic)
  (lambda arguments
    (oo-apply report-meroon-error 'Domain generic
              "No method for" arguments ) ) )

;;; Create an indexed dispatcher. NO is the default dispatcher and CN
;;; is the number of the class on which to index the methods.

(define (create-Indexed-Dispatcher cn . no)
  (let ((cl (number->class cn)))
    (let* ((no (if (pair? no) 
                   (car no) 
                   (make-Immediate-Dispatcher 
                    Immediate-Dispatcher-find-method
                    #f ) ))
           (min (Class-min-son cl))
           (max (Class-max-son cl))
           (sz (+ 1 (- max min)))
           (size (+ sz 1 (quotient sz 2))) ; extra room
           (d ((Class-allocator Indexed-Dispatcher-class) size)) )
      (careless-initialize-Dispatcher-method-finder
       d Indexed-Dispatcher-find-method )
      (set-Indexed-Dispatcher-class-number! d cn)
      (set-Indexed-Dispatcher-min! d min)
      (set-Indexed-Dispatcher-max! d max)
      (set-Indexed-Dispatcher-no! d no)
      (do ((i 0 (+ 1 i)))
          ((>= i size))
        (set-Indexed-Dispatcher-method! d i barfing-method) )
      (initialize! d) ) ) )

;;; Takes an Indexed-Dispatcher and extends it.

(define (extend-Indexed-Dispatcher d)
  (let ((old-size (Indexed-Dispatcher-method-length d))
        (newd (create-Indexed-Dispatcher 
               (Indexed-Dispatcher-class-number d)
               (Indexed-Dispatcher-no d) )))
    (do ((i 0 (+ 1 i)))
        ((= i old-size) newd)
      (set-Indexed-Dispatcher-method! 
       newd i (Indexed-Dispatcher-method d i) ) ) ) )

;;; Create an alternative dispatcher annd fill dependent slots.

(define (create-Subclass-Dispatcher cn no yes)
  (let* ((cl (number->class cn))
         (min (Class-min-son cl))
         (max (Class-max-son cl)) )
    (make-Subclass-Dispatcher 
     Subclass-Dispatcher-find-method
     cn no min max yes ) ) )

;;;============================================================= create generic
;;; Create a new Generic instance, also check version compatibility.
;;; Try to reuse the number associated to a generic having the same
;;; name if any.

(define (register-Generic-1 rev name default-method signature top-class-name)
  (check-revision rev)
  (let ((g (make-Generic-1
            careless-determine-method1  ; behavior
            name                        ; name
            default-method              ; default
            signature                   ; variables
            meroon-uninitialized        ; dispatcher
            meroon-uninitialized ) ))   ; top-classes
    ;; set up a default method if there is none
    (unless (field-defined? g 'default)
      (careless-initialize-Generic-default
       g (create-default-generic-method g) ) )
    ;; set up the top-classes
    (careless-initialize-Generic-top-classes
     g (list (if top-class-name 
                 (symbol->class top-class-name 
                                (lambda (name)
                                  (report-meroon-error
                                   'Syntax 'define-generic
                                   "No such class" name ) ) )
                 #f )) )
    ;; create the dispatcher
    (careless-set-Generic-dispatcher! 
     g (if-meroon-feature renumber-classes ;      Note(*)
           (let ((defdispatcher (make-Immediate-Dispatcher 
                                 Immediate-Dispatcher-find-method
                                 #f ))
                 (top-class (car (careless-Generic-top-classes g))) )
             (if (and (Class? top-class)
                      (not (eq? top-class Object-class)) )
                 (create-Indexed-Dispatcher 
                  (careless-Class-number top-class)
                  defdispatcher )
                 defdispatcher ) )
           (let* ((size (vector-length *classes*))
                  (d ((Class-allocator Global-Dispatcher-class) size)) )
             (careless-initialize-Dispatcher-method-finder
              d Global-Dispatcher-find-method ) 
             d ) ) )
    ;; Propagate default-method to Global-Dispatcher or Indexed-Dispatcher
    (let ((d (careless-Generic-dispatcher g)))
      (when (Global-Dispatcher? d)
        (let ((size (Global-Dispatcher-method-length d))
              (default (careless-Generic-default g)) )
          (do ((i 0 (+ 1 i)))
              ((>= i size))
            (careless-set-Global-Dispatcher-method! d i default) ) ) )
      (when (Indexed-Dispatcher? d)
        (let ((size (Indexed-Dispatcher-method-length d))
              (default (careless-Generic-default g)) )
          (do ((i 0 (+ 1 i)))
              ((>= i size))
            (careless-set-Indexed-Dispatcher-method! d i default) ) ) ) )
    ;; register in the sequence of all generic functions
    (let* ((i (or (look-generic (Generic-name g))
                  (get-new-generic-number) )))
      (vector-set! *generics* i g) )
    ;; generic functions are regular instances so submit them to initialize!
    (initialize! (fill-other-fields! g)) ) )

;;; Note(*): When classes are not renumbered, then generic functions are 
;;; equipped with a global method dispatch vector and find-method is adapted 
;;; to this simpler structure.

(define (register-Generic-N rev name default-method signature . top-class-names)
  (check-revision rev)
  (let ((g (make-Generic-N
            determine-method            ; behavior
            name                        ; name
            default-method              ; default
            signature                   ; variables
            meroon-uninitialized        ; dispatcher
            meroon-uninitialized ) ))   ; top-classes
    ;; set up a default method if there is none
    (unless (field-defined? g 'default)
      (careless-initialize-Generic-default
       g (create-default-generic-method g) ) )
    ;; set up the top-classes
    (careless-initialize-Generic-top-classes
     g (oo-map (lambda (name)
                 (if name (symbol->class name 
                                         (lambda (name)
                                           (report-meroon-error
                                            'Syntax 'define-generic
                                            "No such class" name ) ) )
                     #f ) )
               top-class-names ) )
    ;; create the dispatcher
    (careless-set-Generic-dispatcher! 
     g (make-Immediate-Dispatcher 
        Immediate-Dispatcher-find-multi-method  ; <- multi-method
        #f ) )
    ;; register in the sequence of all generic functions
    (let* ((i (or (look-generic (Generic-name g))
                  (get-new-generic-number) )))
      (vector-set! *generics* i g) )
    ;; generic functions are regular instances so submit them to initialize!
    (initialize! (fill-other-fields! g)) ) )

;;; Find if a previous generic function with the same name is
;;; registered, return its number if it exists.

(define (look-generic name)
  (define (look i)
    (if (< i *generic-number*)
        (if (eq? name (careless-Generic-name (vector-ref *generics* i)))
            i (look (+ 1 i)) )
        #f ) )
  (look 0) )

;;; The behavior of a generic function with only one discriminating
;;; variable. If you cannot find a method, then try to see if the
;;; object has an obsolete class-number, update it and retry to find a
;;; method.  This time if such a method cannot be found then default
;;; to the Generic default.

(define (careless-determine-method1 g o1)
  (if (Object? o1)
      (let ((d (careless-Generic-dispatcher g)))
        (let ((method (find-method1 d (object->class-number o1))))
          (or method
              (if-meroon-feature renumber-classes
                  (and (update-object->class-number! o1)
                       (find-method1 d (object->class-number o1)) )
                  #f )
              (careless-Generic-default g) ) ) )
      (careless-Generic-default g) ) )

;;; The general function to determine a method. O* are the discriminating
;;; instances, since to discriminate on more than one argument is
;;; rare, the search through linear dispatchers expensive, update
;;; class-numbers immediately. Due to the integration with Scheme,
;;; there is a problem since Object is not the top-class ie there are
;;; Scheme values that are not objects. This might be the case of:
;;;         (define-generic (foo (o) (v)))
;;;         (define-method  (foo (o <some-class>>) (v)) ...)
;;; The actual value for v does not need to be an object at all.

(define (determine-method g . o*)
  ;; there is always at least one discriminating variable
  (if (null? (cdr o*))
      (careless-determine-method1 g (car o*))
      (begin
        (for-each (lambda (o) 
                    (when (Object? o) (update-object->class-number! o)) )
                  o* )
        (let ((method (find-multi-method 
                       (careless-Generic-dispatcher g)
                       (oo-map (lambda (o) 
                                 (if (Object? o) (object->class o) #f) )
                               o* ) )))
          (or method
              (careless-Generic-default g) ) ) ) ) )

;;;========================================================== find-method
;;; find-method works for generic functions with a single discriminating
;;; variable. It takes a dispatcher and a class-number then extracts from 
;;; the dispatcher, the appropriate method. Note that no method is
;;; coded as #f so you might have to fetch the default method in the
;;; generic instance if you get such a result.
;;; find-method is not generic since that would be unbearably slow. To
;;; speed up method dispatch.

(define (Immediate-Dispatcher-find-method d cn)
  (careless-Immediate-Dispatcher-method d) )

(define (Subclass-Dispatcher-find-method d cn)
  (if (and (>= cn (careless-Subclass-Dispatcher-min d))
           (<= cn (careless-Subclass-Dispatcher-max d)) )
      (find-method1 (careless-Subclass-Dispatcher-yes d) cn)
      (find-method1 (careless-Subclass-Dispatcher-no d) cn) ) )

(define (Indexed-Dispatcher-find-method d cn)
  (if (and (>= cn (careless-Indexed-Dispatcher-min d))
           (<= cn (careless-Indexed-Dispatcher-max d)) )
      (careless-Indexed-Dispatcher-method 
       d (- cn (careless-Indexed-Dispatcher-min d)) )
      (find-method1 (careless-Indexed-Dispatcher-no d) cn) ) )

(define (Global-Dispatcher-find-method d cn)
  (careless-Global-Dispatcher-method d cn) )

;;; (define (Tracing-Dispatcher-find-method d cn)...) see trace.scm

;;; General method for finding method on Generic-1.

(define (find-method1 d cn)
  ((careless-Dispatcher-method-finder d) d cn) )

;;;========================================================= create method
;;; This function is called at run-time to officially register the method.
;;; It also check the revision version of Meroon. This is just an
;;; entry point that is specialized on the number of discriminating
;;; variables.

(define (register-method generic-name
                         variable-list
                         rev
                         method-maker
                         . discriminating-class-names )
  (check-revision rev)
  (let ((g (symbol->generic generic-name
                            (lambda (name)
                              (report-meroon-error 
                               'Domain 'define-method
                               "No such generic function"
                               generic-name discriminating-class-names
                               variable-list method-maker ) ) ))
        (classes (oo-map (lambda (name)
                           (symbol->class name
                                          (lambda (name)
                                            (report-meroon-error 
                                             'Domain 'define-method
                                             "No such class" 
                                             generic-name name ) ) ) )
                         discriminating-class-names )) )
    (unless (coherent-variables? (Generic-variables g) variable-list)
      (report-meroon-error
       'Syntax 'define-method "Non congruent lambda-lists" 
       (Generic-variables g) variable-list ) )
    ;; only accepts methods that are compatible with the definition of
    ;; the generic function.
    (unless (memq (compare-signatures classes (Generic-top-classes g))
                  '(< =) )
      (report-meroon-error
       'Syntax 'define-method "method prohibited by the generic declaration"
       (Generic-top-classes g) classes ) )
    (case (length discriminating-class-names)
      ((1) (let ((c (car classes)))
             (add-1-method! g c (method-maker g c 'void)) ))
      (else (add-N-method! g classes method-maker)) )
    generic-name ) )

;;; Test the congruence of two variable lists.  They may both contain
;;; a final n-ary (dotted) variable (which cannot be a discriminating
;;; variable (of course)). This predicate is used to test if a
;;; method is compatible with a generic function.

(define (coherent-variables? la lb)
  (if (pair? la)
      (if (pair? lb)
          (and (or ;; similar discriminating variable
                   (and (pair? (car la))
                        (pair? (car lb)) )
                   ;; similar regular variable
                   (and (symbol? (car la))
                        (symbol? (car lb)) ) )
               (coherent-variables? (cdr la) (cdr lb)) )
          #f )
      (or (and (null? la) (null? lb))
          ;; similar dotted variable
          (and (symbol? la) (symbol? lb)) ) ) )

;;;======================================================== update dispatcher
;;; When a class is created then, for each generic function, propagate
;;; the methods of the superclass towards this class. While scanning
;;; dispatchers, update also the copy of the max fields of
;;; Indexed-Dispatchers of Subclass-Dispatchers [Recall that a fresh
;;; class is created with the highest class-number possible (by
;;; renumbering) so the min field is already correct.]  Another case
;;; is possible when class already existed. In that case, just reset
;;; the methods. The dispatchers are supposed to verify that if C1 is
;;; a super of C2 then C1 cannot occur in the no part of a
;;; Subclass-Dispatcher on C2.

;;; This function confers to GENERIC the method of the super of CLASS.

(define-temporary (Generic-update! generic class)
  (cond ((Generic-1? generic) (Generic-1-update! generic class))
        ((Generic-N? generic) (Generic-N-update! generic class))
        (else (report-meroon-error 'internal 'Generic-update!
                                   "No such method" generic class)) ) )

;;; It is not necessary to update generic-N functions since
;;; multimethods use Linear-Dispatcher that do not require to be
;;; updated.

(define (Generic-N-update! generic class)
  generic )

;;; Walk a Generic-1 dispatcher to update Subclass-Dispatcher and
;;; Indexed-Dispatcher ones. update-dispatcher! is a generic function.

(define (Generic-1-update! generic class)
  (set-Generic-dispatcher! 
   generic (update-dispatcher! (Generic-dispatcher generic) class) ) )

(define-temporary (update-dispatcher! d class)
  ((cond ((Immediate-Dispatcher? d) Immediate-Dispatcher-update-dispatcher!)
         ((Subclass-Dispatcher? d)  Subclass-Dispatcher-update-dispatcher!)
         ((Indexed-Dispatcher? d)   Indexed-Dispatcher-update-dispatcher!)
         ((Global-Dispatcher? d)    Global-Dispatcher-update-dispatcher!)
         ((Tracing-Dispatcher? d)   Tracing-Dispatcher-update-dispatcher!)
         (else                      default-update-dispatcher!) )
   d class ) )

;;; should never be called
(define default-update-dispatcher! internal-meroon-error)

(define (Immediate-Dispatcher-update-dispatcher! d class)
  d )

(define (Subclass-Dispatcher-update-dispatcher! d class)
  (let ((cl (number->class (careless-Subclass-Dispatcher-class-number d))))
    (careless-set-Subclass-Dispatcher-max! d (careless-Class-max-son cl))
    (if (careless-subclass? class cl)
        (careless-set-Subclass-Dispatcher-yes! 
         d (update-dispatcher! (careless-Subclass-Dispatcher-yes d) class) )
        (careless-set-Subclass-Dispatcher-no! 
         d (update-dispatcher! (careless-Subclass-Dispatcher-no d) class) ) ) 
    d ) )

(define (Indexed-Dispatcher-update-dispatcher! d class)
  (let* ((cn (careless-Indexed-Dispatcher-class-number d))
         (cl (number->class cn))
         (min (careless-Class-min-son cl))
         (max (careless-Class-max-son cl)) )
    (careless-set-Indexed-Dispatcher-max! d max)
    (let ((d (if (> (careless-Indexed-Dispatcher-method-length d) (- max min))
                 d 
                 (extend-Indexed-Dispatcher d) )))
      (if (careless-subclass? class cl)
          (let ((method (find-method1 d (careless-Class-super-number class)))
                (old-method (find-method1 d (careless-Class-number class))) )
            (define (fill! cn)
              (let ((content (find-method1 d cn)))
                (when (eq? content old-method)
                  (careless-set-Indexed-Dispatcher-method! 
                   d (- cn (careless-Indexed-Dispatcher-min d)) method )
                  (for-each fill! (careless-Class-subclass-numbers 
                                   (number->class cn) )) ) ) )
            (fill! (careless-Class-number class)) )
          ;; walk the `no' part
          (careless-set-Indexed-Dispatcher-no!
           d (update-dispatcher! (careless-Indexed-Dispatcher-no d) class) ) )
      d ) ) )

(define (Global-Dispatcher-update-dispatcher! d class)
  (let ((method (find-method1 d (careless-Class-super-number class)))
        (old-method (find-method1 d (careless-Class-number class))) )
    (define (fill! cn)
      (let ((content (find-method1 d cn)))
        (when (eq? content old-method)
          (careless-set-Global-Dispatcher-method! d cn method)
          (for-each fill! (careless-Class-subclass-numbers 
                           (number->class cn) )) ) ) )
    (fill! (careless-Class-number class))
    d ) )

;;; Traced generic are transparent for addition of classes.

(define (Tracing-Dispatcher-update-dispatcher! d class)
  (set-Tracing-Dispatcher-dispatcher!
   d (update-dispatcher! (Tracing-Dispatcher-dispatcher d) class) )
  d )

;;;======================================================== augment dispatcher
;;; Add a method to a generic function on a particular class. This is
;;; actually done by assigning the generic instance an augmented
;;; dispatcher. It can also be done with a functional style, in this way
;;; a new generic instance would be synthetized rather than modifying
;;; the old one but this consumes a stupendous amount of dispatchers
;;; when a new class is created.

(define (add-1-method! generic class method)
  (let* ((dd (Generic-dispatcher generic))
         (d (augment-dispatcher! dd class method)) )
    (if-meroon-feature renumber-classes
        (set! d (compress-dispatcher! d *dispatcher-max-depth* Object-class))
        #t )
    (set-Generic-dispatcher! generic d)
    (Generic-name generic) ) )

;;; Compute a new dispatcher that provides method on class without
;;; side-effect. Try to compress dispatchers if few methods otherwise
;;; use indexed dispatch tables.

(define-temporary (augment-dispatcher! d class method)
  ((cond ((Immediate-Dispatcher? d) Immediate-Dispatcher-augment-dispatcher!)
         ((Subclass-Dispatcher? d)  Subclass-Dispatcher-augment-dispatcher!)
         ((Indexed-Dispatcher? d)   Indexed-Dispatcher-augment-dispatcher!)
         ((Global-Dispatcher? d)    Global-Dispatcher-augment-dispatcher!)
         ((Tracing-Dispatcher? d)   Tracing-Dispatcher-augment-dispatcher!)
         (else                      default-augment-dispatcher!) )
   d class method ) )

;;; should never be called
(define default-augment-dispatcher! internal-meroon-error)

(define (Immediate-Dispatcher-augment-dispatcher! d class method)
  (cond ((eq? method (Immediate-Dispatcher-method d)) d)
        ;; Often the case:
        ((eq? class Object-class)
         (make-Immediate-Dispatcher 
          Immediate-Dispatcher-find-method
          method ) )
        (else (let ((yes (make-Immediate-Dispatcher 
                          Immediate-Dispatcher-find-method
                          method ))
                    (no d) )
                (create-Subclass-Dispatcher
                 (Class-number class) no yes ) )) ) )

;;; Add a method on a class in an Indexed-Dispatcher. There are two
;;; cases whether the class already exists (then patch the vector
;;; of methods (to avoid copying the Indexed-Dispatcher)) or is
;;; freshly created. In this latter case, a new Indexed-Dispatcher
;;; must be allocated to hold this new method as well as all the old
;;; ones.

(define (Indexed-Dispatcher-augment-dispatcher! d class method)
  (let* ((cn (Indexed-Dispatcher-class-number d))
         (cl (number->class cn)) )
    (set-Indexed-Dispatcher-max! d (Class-max-son cl))
    (let ((d (if (> (Indexed-Dispatcher-method-length d)
                    (- (careless-Class-max-son cl) 
                       (careless-Class-min-son cl) ) )
                 d 
                 (extend-Indexed-Dispatcher d) )))
      (cond 
       ;; also take care of class = cl:
       ((careless-subclass? class cl)
        (let ((old-method (find-method1 d (Class-number class)))
              (min (Indexed-Dispatcher-min d)) )
          (define (fill! cn)
            (let ((content (find-method1 d cn)))
              (when (eq? content old-method)
                (set-Indexed-Dispatcher-method! d (- cn min) method)
                (for-each fill! (Class-subclass-numbers 
                                 (number->class cn) )) ) ) )
          (fill! (Class-number class))
          d ) )
       ;; Add a Subclass-Dispatcher before theIndexed-Dispatcher
       ((careless-subclass? cl class)
        (let ((no-d (Indexed-Dispatcher-no d)))
          (set-Indexed-Dispatcher-no! 
           d (make-Immediate-Dispatcher
              Immediate-Dispatcher-find-method
              method ) )
          (make-Subclass-Dispatcher
           Subclass-Dispatcher-find-method
           (careless-Class-number class)
           no-d
           (careless-Class-max-son class) 
           (careless-Class-min-son class) 
           d ) ) )
       ;; Augment (physically) the no part of this Indexed-Dispatcher
       (else
        (let ((no-d (augment-dispatcher! (Indexed-Dispatcher-no d) 
                                         class method )))
          (set-Indexed-Dispatcher-no! d no-d)
          d ) ) ) ) ) )

(define (Global-Dispatcher-augment-dispatcher! d class method)
  (let ((old-method (find-method1 d (careless-Class-number class))))
    (define (fill! cn)
      (let ((content (find-method1 d cn)))
        (when (eq? content old-method)
          (set-Global-Dispatcher-method! d cn method)
          (for-each fill! (Class-subclass-numbers 
                           (number->class cn) )) ) ) )
    (fill! (careless-Class-number class)) )
  d )

;;; Traced dispatchers are transparent for method addition.

(define (Tracing-Dispatcher-augment-dispatcher! d class method)
  (set-Tracing-Dispatcher-dispatcher!
   d (augment-dispatcher! (Tracing-Dispatcher-dispatcher d) class method) )
  d )

;;; Three cases depending on the relation between CLASS and the class
;;; on which the dispatcher dispatches: CL. Do not forget to update
;;; the copy of max-son in the dispatcher (the min-son never needs to
;;; be updated). One invariant that must be respected is: if C1 is a
;;; superclass of C2 then a Subclass-Dispatcher on C2 cannot have a
;;; dispatcher for C1 in its no part. 

(define (Subclass-Dispatcher-augment-dispatcher! d class method)
  (let ((cl (number->class (Subclass-Dispatcher-class-number d))))
    (set-Subclass-Dispatcher-max! d (Class-max-son cl))
    (cond 
     ;; augment the `yes' part
     ((eq? cl class)
      (let ((yes-d (augment-dispatcher! (Subclass-Dispatcher-yes d) 
                                        class method )))
        (if (and (Subclass-Dispatcher? yes-d)
                 (= (Subclass-Dispatcher-class-number d)
                    (Subclass-Dispatcher-class-number yes-d) ) )
            ;; remove one level
            (set-Subclass-Dispatcher-yes! d (Subclass-Dispatcher-yes yes-d))
            ;; keep it
            (set-Subclass-Dispatcher-yes! d yes-d) )
        d ) )            
     ;; also augment the `yes' part
     ((careless-subclass? class cl)
      (set-Subclass-Dispatcher-yes!
       d (augment-dispatcher! (Subclass-Dispatcher-yes d) class method) )
      d )
     ;; insert a new level of Subclass-Dispatcher. Extract from the no part
     ;; the dispatchers inferior to class.
     ((careless-subclass? cl class)
      (let ((newd (create-Subclass-Dispatcher
                   (Class-number class)
                   (Subclass-Dispatcher-no d)
                   (create-Subclass-Dispatcher
                    (Subclass-Dispatcher-class-number d)
                    (make-Immediate-Dispatcher
                     Immediate-Dispatcher-find-method
                     method )
                    (Subclass-Dispatcher-yes d) ) )))
        (set-Subclass-Dispatcher-no!
         newd (rebalance-dispatcher! newd (Subclass-Dispatcher-no newd)) )
        newd ) )
     ;; augment the `no' part
     (else (set-Subclass-Dispatcher-no! 
            d (augment-dispatcher! (Subclass-Dispatcher-no d) class method) )
           d )) ) )

;;; Note (*): This must avoid the following bug: suppose you have
;;; classes A with subclass B and C.  You define mb as a method for B,
;;; then mc for C, you thus have a Subclass-Dispatcher on B with mb as
;;; yes and in the no part a Subclass-Dispatcher with mc as yes and #f
;;; as no.  Suppose now that you define ma for A then you must not put
;;; mc in the no part of a Subclass-Dispatcher for A nor you can put
;;; ma in the no part of the Subclass-Dispatcher for B. You have to
;;; rebalance the tree of dispatchers.

(define-temporary (rebalance-dispatcher! d dno)
  ((cond 
    ((Subclass-Dispatcher? dno)  Subclass-Dispatcher-rebalance-dispatcher!)
    ((Immediate-Dispatcher? dno) Immediate-Dispatcher-rebalance-dispatcher!)
    ((Indexed-Dispatcher? dno)   Indexed-Dispatcher-rebalance-dispatcher!)
    ((Global-Dispatcher? dno)    Global-Dispatcher-rebalance-dispatcher!)
    ((Tracing-Dispatcher? dno)   Tracing-Dispatcher-rebalance-dispatcher!)
    (else                        internal-meroon-error) )
   d dno ) )

(define (Subclass-Dispatcher-rebalance-dispatcher! d dno)          
  (let ((cl (number->class (Subclass-Dispatcher-class-number d)))
        (clno (number->class (Subclass-Dispatcher-class-number dno))) )
    (if (careless-subclass? clno cl)
        (let ((no (Subclass-Dispatcher-no dno)))
          (set-Subclass-Dispatcher-no!
           dno (Subclass-Dispatcher-yes d) )
          (set-Subclass-Dispatcher-yes! d dno)
          (rebalance-dispatcher! d no) )
        ;; Due to the invariant, clno and cl are unrelated
        (set-Subclass-Dispatcher-no!
         dno (rebalance-dispatcher! 
              d (Subclass-Dispatcher-no dno) ) ) ) ) )

(define (Immediate-Dispatcher-rebalance-dispatcher! d dno)
  dno )

(define (Indexed-Dispatcher-rebalance-dispatcher! d dno)
  (let ((cl (number->class (Subclass-Dispatcher-class-number d)))
        (clno (number->class (Indexed-Dispatcher-class-number dno))) )
    (if (careless-subclass? clno cl)
        (let ((no (Indexed-Dispatcher-no dno)))
          (set-Indexed-Dispatcher-no!
           dno (Subclass-Dispatcher-yes d) )
          (set-Subclass-Dispatcher-yes! d dno)
          (rebalance-dispatcher! d no) )
        ;; Due to the invariant, clno and cl are unrelated
        (set-Indexed-Dispatcher-no!
         dno (rebalance-dispatcher! 
              d (Indexed-Dispatcher-no dno) ) ) ) ) )

(define (Global-Dispatcher-rebalance-dispatcher! d dno)
  dno )

(define (Tracing-Dispatcher-rebalance-dispatcher! d dno)
  dno )

;;;======================================================== compress dispatcher
;;; Convert a dispatcher into an Indexed-Dispatcher if too bushy. This
;;; will speed-up method looking. Compressing dispatchers use LEVEL
;;; to record the depth of the dispatching tree. Whenever too deep,
;;; access is done through an indexing scheme. TOP-CLASS is the
;;; greatest possible class for an instance reaching the current D.

(define-temporary (compress-dispatcher! d level top-class)
  ((cond ((Immediate-Dispatcher? d) Immediate-Dispatcher-compress-dispatcher!)
         ((Indexed-Dispatcher? d)   Indexed-Dispatcher-compress-dispatcher!)
         ((Subclass-Dispatcher? d)  Subclass-Dispatcher-compress-dispatcher!)
         ((Global-Dispatcher? d)    Global-Dispatcher-compress-dispatcher!)
         ((Tracing-Dispatcher? d)   Tracing-Dispatcher-compress-dispatcher!)
         (else                      internal-meroon-error) )
   d level top-class ) )

;;; Converts a dispatcher into an indexed dispatcher. It will be
;;; accessed by all instances of CLASS. 

(define (indexize-dispatcher dispatcher class)
  ;;(display `(indexize-dispatcher ,dispatcher ,class))(newline) ; DEBUG
  (let ((lowest (Class-min-son class))
        (newd (create-Indexed-Dispatcher 
               (Class-number class)
               (make-Immediate-Dispatcher 
                Immediate-Dispatcher-find-method 
                #f ) )) )
    (let fill! ((cn (careless-Class-number class)))
      (set-Indexed-Dispatcher-method! 
       newd (- cn lowest) (find-method1 dispatcher cn) )
      (for-each fill! (Class-subclass-numbers (number->class cn))) ) 
    ;;(display `(result is ,newd))(newline) ; DEBUG
    newd ) )

(define (Global-Dispatcher-compress-dispatcher! d level top-class)
  d )

(define (Immediate-Dispatcher-compress-dispatcher! d level top-class)
  d )

(define (Indexed-Dispatcher-compress-dispatcher! d level top-class)
  d )

(define (Tracing-Dispatcher-compress-dispatcher! d level top-class)
  d )

;;; When compressing a Subclass-Dispatcher and if the dispatcher
;;; corresponding to the NO part is indexed then there is no point to
;;; not indexize also the YES part since they are accessed by the same
;;; TOP-CLASS. This version of compress-dispatcher! is no longer
;;; functional to lessen data consumption.

(define (Subclass-Dispatcher-compress-dispatcher! d level top-class)
  (if (<= level 0) 
      (indexize-dispatcher d top-class)
      (let ((no-d (compress-dispatcher! 
                   (Subclass-Dispatcher-no d) (- level 1) top-class )))
        (if (Indexed-Dispatcher? no-d)
            (indexize-dispatcher d top-class)
            (let ((yes-d (compress-dispatcher! 
                          (Subclass-Dispatcher-yes d)
                          (- level 1)
                          (number->class 
                           (Subclass-Dispatcher-class-number d) ) )))
              (if (and (Indexed-Dispatcher? yes-d)
                       (= (Subclass-Dispatcher-class-number d)
                          (Indexed-Dispatcher-class-number yes-d) ) )
                  ;; remove the Subclass-Dispatcher level
                  (begin 
                    (set-Indexed-Dispatcher-no! yes-d no-d)
                    yes-d )
                  ;; keep the Subclass-Dispatcher level
                  (begin
                    (set-Subclass-Dispatcher-yes! d yes-d)
                    (set-Subclass-Dispatcher-no! d no-d)
                    d ) ) ) ) ) ) )

;;; NOTE: This compaction is too simple-minded, should be improved!!!

;;; end of dispatch.scm
