;;; $Id: renum.scm,v 3.11 1994/09/12 17:31:18 queinnec Exp $
;;; Copyright (c) 1990-93 by Christian Queinnec. All rights reserved.

;;;                ***********************************************
;;;                    Small, Efficient and Innovative Class System
;;;                                    Meroon 
;;;                              Christian Queinnec  
;;;                   \'Ecole Polytechnique & INRIA--Rocquencourt
;;;                ************************************************

;;; This file implements classes renumeration. The goal is to ensure
;;; that two numbers (min-son and max-son) can be associated to any
;;; class such that for any instance of this class, its class-number
;;; lies between these two numbers. An interesting corollary is that
;;; is-a? can be implemented as a two-number interval membership.  A
;;; simple way to achieve this property is to number classes with a
;;; prefix walk of the class tree. In the following picture, the
;;; proper class-number of the class appears wwithin parentheses.

;;;        (0)Object 0-5
;;;           /    \
;;;      (1)Point   \ 
;;;    1-3 / \     Class(4) 
;;;       /   \        \ 4-5 
;;; (2)3DPoint \        \  
;;;     2-2    Rect(3)   \  
;;;             3-3     CountingClass(5)
;;;                         5-5

;;; The problem is that adding new classes, for instance adding a
;;; class such as Colored3DPoint as a subclass of 3DPoint, breaks the
;;; invariant since there is no possible integer number that can be
;;; given to Colored3DPoint. The trick is to renumber the classes but
;;; since many instances might exist and given that we do not want to
;;; force a global GC to renumber them, we will (1) defer this update
;;; until a generic function call and (2) never reuse old
;;; class-numbers. Moreover we try to minimize the number of classes
;;; that are renumbered.

;;; If we are about to add Colored3DPoint as a subclass of 3DPoint, we
;;; therefore renumber all the classes that are to the `left' of
;;; 3DPoint.  This will give:
 
;;;        (6)Object 4-10
;;;           |     \
;;;           |    Point(7)7-10
;;;      (4)Class       |   \
;;;     4-5   |         |    \ 
;;;           |    Rect(8)8-8 \
;;;           |                \
;;;           |               3DPoint(9)9-10
;;;   (5)CountingClass           \
;;;         5-5                   \
;;;                              Colored3DPoint(10)10-10

;;; Observe that now 3DPoint is at the rightmost position so new classes
;;; can be added without requiring to renumber anything. This will be
;;; nevertheless the case if one wants to subclass Rect. Also
;;; observe that branches that are to the right of the new class
;;; (Class and CountingClass) are not renumbered.

;;; The philosophy is to patch in place all objects that contain
;;; class-numbers. These updated fields must then be mutable. The main
;;; problem of these algorithms is that the situation is unstable:
;;; While class-numbers are modified, Indexed-Dispatchers are
;;; problematic since they dispatch on the old number. The generic
;;; machinery ensures that instances with an old class-number are
;;; patched on the fly to get the new class-number but at that time,
;;; it is no more possible to use a not yet updated Indexed-Dispatcher.
;;; The trick is to avoid to use generic functions during this phase.

;;; Show if true some internal details on renumbering
(define *meroon-debug* #f)

;;; This procedure renumbers any class that needs it starting with CLASS. 
;;; This class is the superclass of the future class that will be
;;; created and which needs renumbering. Of course no work is needed
;;; if CLASS is Object.

(define (renumber-classes class)
  (if-meroon-feature renumber-classes
      (unless (eq? class Object-class)
        (let ((rightmost (careless-Class-max-son Object-class))
              (right (careless-Class-max-son class)) )
          ;; renumber classes if CLASS is not at a rightmost position in the
          ;; class hierarchy. 
          (unless (= right rightmost)
            ;; emits a warning so class definitions can be reordered in
            ;; source files to avoid useless renumbering.
            (report-meroon-warning
             'add-subclass "Renumbering classes with respect to" class )
            ;; show class numbers for debug purposes
            (when *meroon-debug*        ; DEBUG
              (show-detailed-hierarchy "before renumbering") )
            ;; renumber ancestors and their proper offspring
            (reorder class)
            ;; renumber the current offspring
            (for-each (lambda (cn)
                        (recursive-renumber! (number->class cn)) )
                      (careless-Class-subclass-numbers class) )
            ;; rebuild min-son and max-son for the class and all its supers
            (reset-son-numbers! class)
            ;; patch all homonymic classes to have an uptodate
            ;; class-number. See note (1).
            (sequence-map
             (lambda (c1)
               (when c1 (patch-homonymic-class c1)) )
             *classes* )
            ;; since min-son and max-son are correct, update dispatchers now.
            (sequence-map (lambda (g) (when g (Generic-renumber! g)))
                          *generics* )
            (when *meroon-debug*        ; DEBUG
              (show-detailed-hierarchy "after renumbering")
              (check-class-consistency Object-class) )
            ) ) )
      #f ) )

;;; Note(1): this pass is necessary for the :prototype option under
;;; interpreter. Accessors close over classes but a new define-class
;;; always register a new one in the *classes* vector. A :prototype
;;; class does not redefine the accessors so the old ones are still
;;; there and they contain a reference (in check-class) to an obsolete
;;; class object. So renumbering patches these old classes to always
;;; have the correct class-number of the definite class. 

(define (patch-homonymic-class c1)
  (let ((c2 (number->class (careless-Class-number c1))))
    (when (and (eq? (careless-Class-name c1) (careless-Class-name c2))
               (not (eq? c1 c2)) )
      ;; setup class number so it is always correct
      (careless-set-Class-number! 
       c1 (careless-Class-number c2) )
      ;; ditto for super number
      (careless-set-Class-super-number!
       c1 (careless-Class-super-number c2) )
      ;; ditto for all supers since this is used in subclass?
      (let ((path-length (careless-Class-super-length c1)))
        (do ((i 0 (+ i 1)))
            ((>= i path-length))
          (careless-set-Class-super! 
           c1 i (careless-Class-super c2 i) ) ) ) ) ) )

;;; Move a class, all its supers and siblings. At the top of hierarchy
;;; only move the classes that are at the left; in between move all
;;; classes except the one which needs to be rightmost; at the bottom
;;; move all offspring. One cannot rebuild min-son and max-son here since
;;; we must wait for the whole offspring to be moved.

(define (reorder class)
  (let ((super (Class-super-class class)))
    (if (eq? super Object-class)
        (begin 
          ;; only move Object-class if its class-number is too low
          (when (< (careless-Class-number Object-class) 
                   (careless-Class-number class) )
            (Class-renumber! Object-class)
            (update-object-class-number! 
             Object-class (careless-Class-number Object-class) ) )
          (for-each (lambda (cn)
                      (let ((c (number->class cn)))
                        (cond 
                         ((eq? c class) 'nothing)
                         ;; move only classes that are at the left of CLASS
                         ((< (careless-Class-max-son c) 
                             (careless-Class-min-son class) )
                          (recursive-renumber! c) ) ) ) )
                    (careless-Class-subclass-numbers Object-class) ) )
        ;; Normal case
        (begin
          (reorder (Class-super-class class))
          ;; Move all other siblings
          (for-each (lambda (cn)
                      (let ((c (number->class cn)))
                        (unless (eq? c class)
                          (recursive-renumber! c) ) ) )
                    (careless-Class-subclass-numbers super) ) ) )
    ;; and move it rightmost
    (Class-renumber! class) ) )

;;; Update the class number of Object in every class whether moved or not.

(define (update-object-class-number! c object-cn)
  (when (and (careless-Class-super-number c)
             (eq? (number->class (careless-Class-super-number c))
                  Object-class ) )
    (careless-set-Class-super-number! c object-cn) )
  (careless-set-Class-super! c 0 object-cn)
  (for-each (lambda (cn) 
              (update-object-class-number! (number->class cn) object-cn) )
            (careless-Class-subclass-numbers c) ) )

;;; Renumber the class and all its offspring. The class that must be moved
;;; to the rightmost place does not appear in the offspring.

(define (recursive-renumber! class)
  (Class-renumber! class)
  (for-each (lambda (cn) (recursive-renumber! (number->class cn)))
            (careless-Class-subclass-numbers class) )
  (set-class-son-numbers! class) )

;;; These functions cannot be method for stability problems.

;;; Renumber a Class instance, do not recurse on subclasses.

(define (Class-renumber! o)
  (unless (eq? o Object-class)
    ;; patch the super-number. Since renumbering starts from Object,
    ;; the class-number of the super-class is now accurate.
    (careless-set-Class-super-number!
     o (current-class-number (careless-Class-super-number o)) ) )
  ;; Move the class 
  (let ((ocn (careless-Class-number o))
        (ncn (get-new-class-number)) )
    (vector-set! *classes* ncn o)
    (careless-set-Class-number! o ncn)
    ;; The current class-number is now correct so update all super numbers.
    ;; DANGER: when modifying Class-class, the object has a temporarily
    ;; inconsistant class-number which may clash with is-a?
    ;; (update-object->class-number! o)
    (let ((n (careless-Class-super-length o)))
      (do ((i 0 (+ 1 i)))
          ((>= i n))
        (careless-set-Class-super! 
         o i (current-class-number (careless-Class-super o i)) ) ) )
    ;; an update fields
    (for-each Field-renumber! (careless-Class-fields o))
    ;; let the super knows the new number (ncn) instead of the old one (ocn)
    (unless (eq? o Object-class)
      (let patch ((cns (careless-Class-subclass-numbers (Class-super-class o))))
        (when (pair? cns)
          (if (= ocn (car cns))
              (set-car! cns ncn)
              (patch (cdr cns)) ) ) ) ) ) )

;;; Fields are simply renumbered. Just keep track of the new number of
;;; the class that introduces them.

(define (Field-renumber! f)
  (careless-set-Field-class-number! 
   f (current-class-number (careless-Field-class-number f)) ) )

;;; Generic functions have dispatchers that use quite heavily class-numbers.
;;; Recurse on them.

(define (Generic-renumber! o)
  (Dispatcher-renumber! (careless-Generic-dispatcher o)) )

;;; Handle all sort of Fields. This is not a generic function since
;;; while renumbering, dispatchers are not stable.
;;; FUTURE: will disappear with the new dispatch mechanism of Meroon4.

(define (Dispatcher-renumber! o)
  (let ((mf (careless-Dispatcher-method-finder o)))
    (cond ((eq? mf Immediate-Dispatcher-find-method)
           (Immediate-Dispatcher-renumber! o) )
          ((eq? mf Subclass-Dispatcher-find-method)
           (Subclass-Dispatcher-renumber! o) )
          ((eq? mf Indexed-Dispatcher-find-method)
           (Indexed-Dispatcher-renumber! o) )
          ((eq? mf Linear-Dispatcher-find-multi-method)
           (Linear-Dispatcher-renumber! o) )
          ((eq? mf Immediate-Dispatcher-find-multi-method)
           (Immediate-Dispatcher-renumber! o) )
          ((eq? mf Global-Dispatcher-find-method)
           (Global-Dispatcher-renumber! o) )
          (else (internal-meroon-error)) ) ) )

(define (Global-Dispatcher-renumber! o) o)

(define (Immediate-Dispatcher-renumber! o) o)

;;; Linear-Dispatchers presently refer to Class instance so they do
;;; not need to be patched: this is automatic.

(define (Linear-Dispatcher-renumber! o) o)

(define (Subclass-Dispatcher-renumber! o)
  (let* ((cn (current-class-number 
              (careless-Subclass-Dispatcher-class-number o) ))
         (class (number->class cn)) )
    (careless-set-Subclass-Dispatcher-class-number! o cn)
    (careless-set-Subclass-Dispatcher-min! o (careless-Class-min-son class))
    (careless-set-Subclass-Dispatcher-max! o (careless-Class-max-son class))
    (Dispatcher-renumber! (careless-Subclass-Dispatcher-no o))
    (Dispatcher-renumber! (careless-Subclass-Dispatcher-yes o)) ) )

;;; Since classes have been renumbered the order of methods must be
;;; restaured. This is a costly operation. One problem is with
;;; stillborn classes which have number but disappear in the
;;; renumbering: their methods must be removed. In fact the semantics
;;; of a disappearing class is not addressed at all.

(define (Indexed-Dispatcher-renumber! o)
  (let* ((oldd (instance-clone o))
         (start (careless-Indexed-Dispatcher-min oldd))
         (stop (careless-Indexed-Dispatcher-max oldd))
         (cn (current-class-number 
              (careless-Indexed-Dispatcher-class-number o) ))
         (class (number->class cn))
         (newmin (careless-Class-min-son class)) )
    (careless-set-Indexed-Dispatcher-class-number! o cn)
    (careless-set-Indexed-Dispatcher-min! o newmin)
    (careless-set-Indexed-Dispatcher-max! o (careless-Class-max-son class))
    ;; reorder methods inside the dispatcher
    (do ((i start (+ 1 i)))
        ((> i stop))
      (let ((ccn (current-class-number i)))
        ;; but only if the class still exists
        (when (and (>= ccn (careless-Class-min-son Object-class))
                   (<= ccn (careless-Class-max-son Object-class)) )
          (careless-set-Indexed-Dispatcher-method! 
           o (- ccn newmin)
           (careless-Indexed-Dispatcher-method oldd (- i start)) ) ) ) )
    (Dispatcher-renumber! (careless-Indexed-Dispatcher-no o)) ) )

;;; Notes:
;;; On pourrait aussi ne pas bouger Object car apres tout tout objet
;;; Meroon appartient a Object. Le defaut est alors que les vecteurs
;;; de dispatch qui discriminent sur toutes les classes sont aussi
;;; grands que *classes*.

;;; end of renum.scm
